require "stats"
-- VARIABLE DECLARATION

TICKS_PER_UPDATE = 15 
-- each reactor and cooling tower gets updated once every 15 ticks (60 ticks = 1 s)
CHANGE_MULTIPLIER = 0.2 
-- used to multiply the temperature change 
-- CHANGE_MULTIPLIER and TICKS_PER_UPDATE work together and are balanced: 
-- 0.2 CHANGE_MULTIPLIER = 15 TICKS_PER_UPDATE
REACTOR_MASS = 4000 
-- used to calculate temperature changes when emergency cooling is used
-- the mass is an estimate best guess based on many tries and errors 

POWER_USAGE_STARTING=2000000 -- 3600 KW
POWER_USAGE_INTERFACE=200000 -- 200 KW
POWER_USAGE_COOLING=1000000 -- 1 MW when fluid amount changed (static) or 1 MW per 20 MW cooling (non-static)
-- entity names
REACTOR_ENTITY_NAME = "realistic-reactor-normal"
REACTOR_POWER_NAME = "realistic-reactor-power-normal"
BREEDER_ENTITY_NAME = "realistic-reactor-breeder"
BREEDER_POWER_NAME = "realistic-reactor-power-breeder"
--REACTOR_ENTITY_NAME_START = "realistic-reactor-start"
--REACTOR_ENTITY_NAME_STOPPED = "realistic-reactor"
--REACTOR_ENTITY_NAME_SCRAMED = "realistic-reactor-scramed"
INTERFACE_ENTITY_NAME = "realistic-reactor-interface"
BREEDER_INTERFACE_ENTITY_NAME = "realistic-breeder-interface"
BOILER_ENTITY_NAME = "realistic-reactor-eccs"
TOWER_ENTITY_NAME = "rr-cooling-tower"
STEAM_ENTITY_NAME = "rr-cooling-tower-steam"
HEAT_PIPE_ENTITY_NAME = "heat-pipe"
--signal names
SIGNAL_CORE_TEMP = {type="virtual", name="signal-reactor-core-temp"}
SIGNAL_STATE_STOPPED = {type="virtual", name="signal-state-stopped"}
SIGNAL_STATE_STARTING = {type="virtual", name="signal-state-starting"}
SIGNAL_STATE_RUNNING = {type="virtual", name="signal-state-running"}
SIGNAL_STATE_SCRAMED = {type="virtual", name="signal-state-scramed"}
SIGNAL_CONTROL_START = {type="virtual", name="signal-control-start"}
SIGNAL_CONTROL_SCRAM = {type="virtual", name="signal-control-scram"}
SIGNAL_COOLANT_AMOUNT = {type="virtual", name="signal-coolant-amount"}
--SIGNAL_COOLANT_TEMP = {type="virtual", name="signal-coolant-temperature"}
SIGNAL_URANIUM_FUEL_CELLS = {type="virtual", name="signal-uranium-fuel-cells"}
SIGNAL_USED_URANIUM_FUEL_CELLS = {type="virtual", name="signal-used-uranium-fuel-cells"}
SIGNAL_REACTOR_POWER_OUTPUT = {type="virtual", name="signal-reactor-power-output"}
SIGNAL_REACTOR_EFFICIENCY = {type="virtual", name="signal-reactor-efficiency"}
SIGNAL_REACTOR_CELL_BONUS = {type="virtual", name="signal-reactor-cell-bonus"}
SIGNAL_REACTOR_ELECTRIC_POWER = {type="virtual", name="signal-reactor-electric-power"}

--explosion sound effect
script.on_event(defines.events.on_trigger_created_entity, function(event)
    local ent = event.entity
    if ent.name == "RR-uranium-explosion-LUQ" then -- this is the left upper quarter of the explosion graphics
        for i, player in pairs(game.connected_players) do
            player.surface.create_entity({name = "RR-nuclear-detonation-sound", position = player.position})
		end
	end
end)


-- sound for radiation damage
script.on_event(defines.events.on_entity_damaged or 97,function(event)
	if ((event.original_damage_amount  >= 0.29999 and event.original_damage_amount < 0.300001) or
		(event.original_damage_amount  >= 0.98999 and event.original_damage_amount < 0.990001)) and 
		event.entity.name == "player" then
		
		if 	global.geigers[event.entity.player.index] == nil or 
			global.geigers[event.entity.player.index]+3 < event.tick then
			
			local player = event.entity
			local rnd = math.floor(math.random()*1.99)
			global.geigers[player.player.index] = event.tick
			player.surface.create_entity({name = "RR-geiger-"..rnd, position = player.position})
		end
	end
end)

-- calculate power output, efficiency, bonus cell production
function calculate_stats_ingo(reactor,state_running_time)
	local temperature = reactor.core.temperature
	local reactor_neighbours = reactor.neighbours
	--local reactor_displayer_name = reactor.displayer.name
	local power 
	local power_breeder
	local efficiency
	local BonusCellAmount
	
	
	-- if reactor.state  == "starting" then
		-- reactor.signals.parameters["state_starting"].count = settings.global["reactor-starting-duration"].value - running_time + 1 --ToDo: starting phase should end, when first fuel cell is half empty
		-- reactor_parameters.power = math.floor(reactor_parameters.power * ((running_time)/settings.global["reactor-starting-duration"].value)^2)
		-- --reactor_parameters.efficiency = reactor_parameters.efficiency
		-- reactor_parameters.bonus_cells = 0
	-- end
	-- if reactor.state  == "scramed" then
		-- reactor.signals.parameters["state_scramed"].count = settings.global["reactor-scram-duration"].value - running_time + 1
		-- reactor_parameters.power = math.floor(reactor_parameters.power * ((settings.global["reactor-scram-duration"].value - (running_time)/3.5)/settings.global["reactor-scram-duration"].value)^11+0.45)
		-- --reactor_parameters.efficiency = reactor_parameters.efficiency
		-- reactor_parameters.bonus_cells = 0
	-- end

	--set or calculate values
	if reactor.state == "running" and temperature<500 then
		power=40
		power_breeder=40
		efficiency=100
		BonusCellAmount=0
	elseif reactor.state == "starting" then
		power=math.min(((39/settings.global["reactor-starting-duration"].value)*state_running_time*2)+1,40) -- running time: 0s=1MW, start_duration/2(default:30s/2)=40MW 
		power_breeder=math.min(((39/settings.global["reactor-starting-duration"].value)*state_running_time*2)+1,40) -- running time: 0s=1MW, start_duration/2(default:30s/2)=40MW
		efficiency=100
		BonusCellAmount=0
	elseif reactor.state == "scramed" then
		power=(((reactor.power_output_last_tick-1)/settings.global["reactor-scram-duration"].value)*(settings.global["reactor-scram-duration"].value-state_running_time))+1 --running time: scram_duration(default:180s)=power_output_last_tick, 0s=1MW
		power_breeder=(((reactor.power_output_last_tick-1)/settings.global["reactor-scram-duration"].value)*(settings.global["reactor-scram-duration"].value-state_running_time))+1 --running time: scram_duration(default:180s)=power_output_last_tick, 0s=1MW
		efficiency=200
		BonusCellAmount=0
	else
	
		--set output
		power = math.max(((2/9)*temperature)-(640/9),40) -- 500°=40MW, 950°=140MW
		power_breeder = math.max(((4/45)*temperature)-(40/9),40) --500°=40MW, 950°=80MW
		BonusCellAmount = math.max(((1/450)*temperature)-(10/9),0) --500°=0 Cells, 950°=1 Cell
		
		--set efficiency
		if reactor_neighbours==1 then
			--single reactor
			if power<=80 then
				efficiency = ((power-40)*(10/4))+100
			else
				efficiency = 200-((power-80)*(10/4))
			end
			
		elseif reactor_neighbours==2 then
			--two reactors
			if power<=80 then
				efficiency = ((power-40)*(10/4))+100
			elseif power>80 and power<=100 then
				efficiency = 200
			else
				efficiency = 200-((power-100)*(10/6))
			end
			
		elseif reactor_neighbours==3 then
			--three reactors
			if power<=80 then
				efficiency = ((power-40)*(10/4))+100
			elseif power>80 and power<=120 then
				efficiency = 200
			else
				efficiency = 200-((power-120)*(10/8))
			end			
			
		else
			--four reactors
			if power<=80 then
				efficiency = ((power-40)*(10/4))+100
			else
				efficiency = 200
			end
		end	
	
	end
	
	
	-- return values	
	if reactor.displayer.name == REACTOR_ENTITY_NAME then	
		return {power = math.floor(power), efficiency = efficiency, bonus_cells = 0, max_power = 151, max_efficiency = 200}
	end
	
	if reactor.displayer.name == BREEDER_ENTITY_NAME then 
		return {power = math.floor(power_breeder), efficiency = efficiency, bonus_cells = BonusCellAmount, max_power = 84, max_efficiency = 200}
	end
	
end


function calculate_stats_ownly(reactor,running_time)
	local fuel_cell = reactor.core.burner.currently_burning
	local temperature = reactor.core.temperature
	local reactors = reactor.neighbours
	reactors = math.min(4,reactors)
	reactors = math.max(1,reactors)
	local efficiency
	local output
	local min_efficiency = 1
	local max_efficiency = 200
	local max_output = 150
	local bonus_cells = 0
	local bonus_cells_multiplier = 1
	temperature = math.max(0,temperature - 500)
	-- https://rechneronline.de/funktionsgraphen/
	

	if reactor.core.burner.currently_burning == nil then
		efficiency=0
		output=0
		
	--SigmaOne's Mods: Nuclear fuel (same stats as uranium power)
	elseif fuel_cell.name ==  "mox-fuel-cell" and reactor.core.burner.currently_burning.fuel_value == 8500000000 then --mox
		efficiency = (3+temperature*0.11)^1.18+80
		output = (1+temperature*0.16)^1.08+35
		min_efficiency = 84
		max_efficiency = 200
		max_output = 150
		bonus_cells_multiplier = 0.66
		
	--Plutonium Energy
	elseif fuel_cell.name == "plutonium-fuel-cell" then --plutonium
		output = (2.2+temperature*0.034)^0.57*25+10
		efficiency = (2.5+temperature*0.034)^0.45*38+45
		min_efficiency = 102
		max_efficiency = 190
		max_output = 145
		
	--Nuclear Fuel
	elseif fuel_cell.name == "breeder-fuel-cell" then --probably a mox cell
		--output = (2.2+temperature*0.034)^0.57*25+10 
		--efficiency = (2.5+temperature*0.034)^0.45*38+45 --these were plutonium stats
		--min_efficiency = 102
		--max_efficiency = 190
		--max_output = 145
		efficiency = (3+temperature*0.11)^1.18+80
		output = (1+temperature*0.16)^1.08+35
		min_efficiency = 84
		max_efficiency = 200
		max_output = 150
		bonus_cells_multiplier = 0.5
		
	--Uranium Power
	elseif fuel_cell.name == "fuel-assembly-mox" then --mox
		efficiency = (3+temperature*0.11)^1.18+80
		output = (1+temperature*0.16)^1.08+35
		min_efficiency = 84
		max_efficiency = 200
		max_output = 150
	
	--Advanced Atomics
	elseif fuel_cell.name == "mox-fuel-cell" then --mox
		efficiency = (3+temperature*0.11)^1.18+80
		output = (1+temperature*0.16)^1.08+35
		min_efficiency = 84
		max_efficiency = 200
		max_output = 150
		bonus_cells_multiplier = 1.45
		
	--Default/other fuel cells 
	else --uranium
		efficiency=(800+temperature*5)^0.64*0.9+29
		output=(1+temperature*0.07)^1.15*1.52+45
		min_efficiency = 94
		max_efficiency = 190
		max_output = 138
		if game.active_mods["Nuclear Fuel"] then
			bonus_cells_multiplier = 0.2
		else
			bonus_cells_multiplier = 0.8
		end
	end
	--breeder
	if reactor.displayer.name == BREEDER_ENTITY_NAME then
		bonus_cells = (0.15+(efficiency-min_efficiency)/(max_efficiency-min_efficiency)*0.85)*bonus_cells_multiplier --15% - 100% *multiplier
		output = output /(1+(efficiency-min_efficiency)/(max_efficiency-min_efficiency)*0.85)
		max_output = max_output /(1+(max_efficiency-min_efficiency)/(max_efficiency-min_efficiency)*0.85)
		efficiency = efficiency/(1.15+(efficiency-min_efficiency)/(max_efficiency-min_efficiency)*0.85)
		max_efficiency = max_efficiency/(1.15+(max_efficiency-min_efficiency)/(max_efficiency-min_efficiency)*0.85)
	end
	--reactor count modifier
	efficiency = efficiency*(0.7+0.075*reactors)
	max_efficiency = math.floor(max_efficiency*(0.7+0.075*reactors))
	
	
	if reactor.state  == "starting" then
		output = math.floor(output * ((running_time)/settings.global["reactor-starting-duration"].value)^2)
		bonus_cells = bonus_cells * ((running_time)/settings.global["reactor-starting-duration"].value)^2
	end
	if reactor.state  == "scramed" then
		output = math.floor(reactor.power_output_last_tick * ((settings.global["reactor-scram-duration"].value - (running_time)/3.5)/settings.global["reactor-scram-duration"].value)^11+0.45)
		bonus_cells = bonus_cells * ((settings.global["reactor-scram-duration"].value - (running_time)/3.5)/settings.global["reactor-scram-duration"].value)^11
	end	
	
	--reactor count modifier ---
	--meine formel
	--local optimal_temp = 90+reactors*90 --+500
	--local remaining_temp_range
	--if reactors >=4 then
	--	remaining_temp_range = 9999999
	--else 
	--	remaining_temp_range = (500-optimal_temp)
	--end
	--if temperature <=optimal_temp then
	--	efficiency = efficiency * (1+0.5*temperature/optimal_temp)
	--else
	--	efficiency = efficiency * (1.5-0.5*(temperature-optimal_temp)/remaining_temp_range)
	--end
	
	
	--angepasst an deine formel, 1R fällt auf 75% ab (0.5*(3/2)), 2R fällt auf 116.66% ab(0.5*(2/3)), 3R fällt auf 137.5% ab (0.5*(1/4))
	--local optimal_temp = 90+reactors*90 --+500
	--local remaining_temp_range
	--if reactors >=4 then
	--	remaining_temp_range = 9999999
	--else 
	--	remaining_temp_range = (500-optimal_temp)
	--end
	--if temperature <=optimal_temp then
	--	efficiency = efficiency * (1+0.5*temperature/optimal_temp)
	--else
	--	efficiency = efficiency * (1.5-0.5*((4-reactors)/(1+reactors))*(temperature-optimal_temp)/remaining_temp_range)
	--end

	if reactor.displayer.name == REACTOR_ENTITY_NAME then	
		return {power=math.floor(output),efficiency = efficiency, bonus_cells = 0, max_power = max_power, max_efficiency = max_efficiency}
	else
		return {power=math.floor(output),efficiency = efficiency, bonus_cells = bonus_cells, max_power = max_output, max_efficiency = max_efficiency}
	end
end




-- INITIALIZING AND UPDATING FUNCTIONS
do


-- mod initialization
function init_global()
  global = global or {}
  -- global.reactors stores the reactor and its parts(core, circuit interface, eccs)
  global.reactors = global.reactors or {}
  -- global.towers stores the cooling tower and the steam maker entity
  global.towers = global.towers or {}
  -- used to store values during build_connected_reactors_list
  global.connected_reactors = {}
  global.all_heat_pipes = {}
  
  game.write_file("RealisticReactors.log"," ") -- this line cleans the log file on game start
  logging("----------------------------------------------------------------------------")
  logging("Realistic Reactors Mod initialized")
  logging("---")
  logging("This is the log file of the Realistic Reactors mod. If you want to get rid of it, set WRITE LOG setting = false in the game options menu.")
  logging("Please note, that this file is deleted and recreated, every time you restart the game.")
  logging("----------------------------------------------------------------------------")
  
  gui_init()
  global.fallout = {}
  global.falloutclouds = {}
  global.geigers = {}
  global.version=1
  
end

script.on_init(init_global)

-- migration function
script.on_configuration_changed(function()

	--game.players[1].print("on_configuration_changed. global.version: "..global.version)

	if not global.version or global.version < 1 then
		for i,reactor in pairs(global.reactors) do
			reactor.signals.parameters["coolant-temperature"] = nil
			reactor.signals.parameters["neighbour-bonus"] = nil
			--fluid signal removal
			--reactor.signals.parameters["coolant-amount"] = nil
			
			if not reactor.efficiency then
				reactor.efficiency=100
			end
			if not reactor.signals.parameters["power-output"] then
				reactor.signals.parameters["power-output"]={signal=SIGNAL_REACTOR_POWER_OUTPUT,count=0,index=7}
			end
			if not reactor.signals.parameters["efficiency"] then
				reactor.signals.parameters["efficiency"]={signal=SIGNAL_REACTOR_EFFICIENCY,count=0,index=10}
			end
			
			--2
			reactor.displayer = reactor.core.surface.create_entity{name = "realistic-reactor-normal", position = {reactor.core.position.x, reactor.core.position.y}, force = reactor.core.force.name}
			if not reactor.core.get_fuel_inventory().is_empty() then
                reactor.displayer.get_fuel_inventory().insert(reactor.core.get_fuel_inventory()[1])
            end
			
			--3
			reactor.fuel_last_tick = reactor.core.burner.remaining_burning_fuel
			if reactor.core.get_fuel_inventory().is_empty() then
				reactor.cells_last_tick = 0
			else
				reactor.cells_last_tick = reactor.core.get_fuel_inventory()[1].count
			end
			reactor.displayer.active = reactor.core.active
			reactor.displayer.minable = reactor.core.minable
			
			--4
			reactor.power = reactor.core.surface.create_entity
            {
            name = REACTOR_POWER_NAME,
            position = reactor.core.position,
            force = reactor.core.force
            }
            reactor.power.destructible=false
			reactor.power.energy = 17000000
            reactor.signals.parameters["coolant-amount"] = {signal=SIGNAL_COOLANT_AMOUNT, count=0, index=6}
			
			--5 (removed, was the old implementation of bonus cells)
			--6
			reactor.signals.parameters["cell-bonus"] = {signal=SIGNAL_REACTOR_CELL_BONUS, count=0, index=11}
			
			--7
			reactor.interface.operable = true
			
			--8
			reactor.cooling_history = 0
			local color = "black"
			if reactor.state == "starting" then
				color = "yellow"
			elseif reactor.state == "running" then
				color = "green"
			elseif reactor.state == "scramed" then
				color = "red"
			end
			reactor.lamp = reactor.core.surface.create_entity{name = "rr-"..color.."-lamp", position = {reactor.core.position.x-0.62,reactor.core.position.y+0.6}, force = reactor.core.force.name}
			reactor.lamp.destructible = false
			reactor.light = reactor.core.surface.create_entity{name = "rr-"..color.."-light", position = {reactor.core.position.x-0.62,reactor.core.position.y+0.6}, force = reactor.core.force.name}
			reactor.light.destructible = false
			reactor.interface_warning_tick = 0
			reactor.interface_warning = nil
			reactor.cooling_warning_tick = 0
			reactor.cooling_warning = nil
			reactor.power_usage = {starting = 0,cooling = 0, interface = 0}
			reactor.neighbours = 0
			reactor.bonus_cells = {}
			reactor.core.destructible = false
			
			--9
			reactor.signals.parameters["electric-power"] = {signal=SIGNAL_REACTOR_ELECTRIC_POWER, count=0, index=12}
			reactor.fluid_amount_last_tick = 0
			reactor.power_output_last_tick = 0
			if reactor.core.get_burnt_result_inventory().is_empty() then
				reactor.used_cells_last_tick = 0
			else
				reactor.used_cells_last_tick = reactor.core.get_burnt_result_inventory()[1].count
			end
			
			
		end
		--7
		gui_init()
		
		--9
		global.fallout = {}
		global.falloutclouds = {}
		global.geigers = {}
		
		global.version=1
	end
	
	for key, force in pairs (game.forces) do
		if force.technologies["nuclear-power"].researched == true then
			force.recipes["nuclear-reactor"].enabled = not settings.startup["disable-vanilla-reactor"].value
			force.recipes["realistic-reactor"].enabled = true
			force.recipes["rr-cooling-tower"].enabled = true
		end
	end
end)

--[[
script.on_event(defines.events.on_player_configured_blueprint, function(event)
	game.players[1].print("on_player_configured_blueprint")
	
	-- access blueprint in players hand
	local stack = game.players[1].cursor_stack
	if stack.valid_for_read then
		
		-- get entities in blueprint
		local ent = stack.get_blueprint_entities()
		
		for i,entity in pairs(ent) do
			
			if string.sub(entity.name,1,17) == "realistic-reactor" then
				game.players[1].print(entity.name.." --> REACTOR")
				game.players[1].print(entity.position)
			end
		end
		
	else
		game.players[1].print("stack not valid_for_read")
	end

end)
]]

-- runs the update functions every x ticks
script.on_event(defines.events.on_tick, function(event)
		if event.tick % 200 ==0 then
			for key,tbl in pairs(global.fallout) do --fallout clouds 80 secs
				if tbl.tick+60*settings.global["clouds_generation"].value < game.tick then
					global.fallout[key] = nil
				else
					local cloud = tbl.surface.create_entity({
						name="fallout-cloud-periodic",
						position=tbl.position
					})	
					table.insert(global.falloutclouds, cloud)
				end
			end

		end
		if event.tick % 79 ==0 then
			if #global.falloutclouds > 0 then
				local rnd = math.random(#global.falloutclouds)
			--for _,cloud in pairs(global.falloutclouds) do --fallout 300 secs
				if global.falloutclouds[rnd] and global.falloutclouds[rnd].valid then
					global.falloutclouds[rnd].surface.create_entity({
						name="permanent-radiation",
						position=global.falloutclouds[rnd].position
					})
				else
					table.remove(global.falloutclouds,rnd)
				end
			end
			--end
		end
	if table_length(global.reactors) > 0 then
	-- run events only when a reactor exists
		--fallout

		--sync reactor inventories
		if event.tick % 11 == 0 then
			for _,reactor in pairs(global.reactors) do
				
				-- storing variables for better performance because they only need to be read once
				local displayer_fuel 
				if reactor.displayer.get_fuel_inventory().is_empty() then
					displayer_fuel = 0
				else
					displayer_fuel = reactor.displayer.get_fuel_inventory()[1].count
				end
				
				local core_fuel
				if reactor.core.get_fuel_inventory().is_empty() then
					core_fuel = 0
				else
					core_fuel = reactor.core.get_fuel_inventory()[1].count
				end
				
				local displayer_burnt
				if reactor.displayer.get_burnt_result_inventory().is_empty() then
					displayer_burnt = 0
				else
					displayer_burnt = reactor.displayer.get_burnt_result_inventory()[1].count
				end
				
				local core_burnt
				if reactor.core.get_burnt_result_inventory().is_empty() then
					core_burnt = 0
				else
					core_burnt = reactor.core.get_burnt_result_inventory()[1].count
				end
				--
				
				
				--sync currently burning
				reactor.displayer.burner.currently_burning = reactor.core.burner.currently_burning
				if reactor.core.burner.remaining_burning_fuel == 0 then
					reactor.displayer.burner.remaining_burning_fuel = 0
				else
					reactor.displayer.burner.remaining_burning_fuel = math.max(20,reactor.core.burner.remaining_burning_fuel)
				end
				
				--if reactor.debug_start_cell and displayer_fuel > 0 then
				--	reactor.cells_last_tick = reactor.cells_last_tick - 1
				--	reactor.debug_start_cell = nil
				--end
				-- sync used fuel cells in core to manipulations in the displayer
				if core_fuel == 0 then
					if displayer_fuel > reactor.cells_last_tick then --used fuel cells have been added to displayer
						reactor.core.get_fuel_inventory().insert(reactor.displayer.get_fuel_inventory()[1])
						reactor.core.get_fuel_inventory()[1].count = displayer_fuel - reactor.cells_last_tick
						core_fuel = displayer_fuel - reactor.cells_last_tick
					end
				else --core has used fuel cells
					if core_fuel + (displayer_fuel - reactor.cells_last_tick) > 0 then --at least 1 cell would remain
						reactor.core.get_fuel_inventory()[1].count = core_fuel + (displayer_fuel - reactor.cells_last_tick)
						core_fuel = core_fuel + (displayer_fuel - reactor.cells_last_tick)
					else --new amount would be 0 or below
						reactor.core.get_fuel_inventory().clear()
						core_fuel = 0
					end
				end 
				
				--sync displayer to new amount in core
				if core_fuel == 0 then
					reactor.displayer.get_fuel_inventory().clear() --sync displayer to new amount in core
					reactor.cells_last_tick = 0 --store amount for next tick
				else
					if displayer_fuel == 0 then --safety check
						reactor.displayer.get_fuel_inventory().insert(reactor.core.get_fuel_inventory()[1]) --sync displayer to new amount in core
					elseif displayer_fuel ~= core_fuel then
						reactor.displayer.get_fuel_inventory()[1].count = core_fuel --sync displayer to new amount in core
					end
					reactor.cells_last_tick = core_fuel --store amount for next tick
				end

				-- sync used fuel cells in core to manipulations in the displayer
				if core_burnt == 0 then
					if displayer_burnt > reactor.used_cells_last_tick then --used fuel cells have been added to displayer
						reactor.core.get_burnt_result_inventory().insert(reactor.displayer.get_burnt_result_inventory()[1])
						reactor.core.get_burnt_result_inventory()[1].count = displayer_burnt - reactor.used_cells_last_tick
						core_burnt = displayer_burnt - reactor.used_cells_last_tick
					end
				else --core has used fuel cells
					if core_burnt + (displayer_burnt - reactor.used_cells_last_tick) > 0 then --at least 1 cell would remain
						reactor.core.get_burnt_result_inventory()[1].count = core_burnt + (displayer_burnt - reactor.used_cells_last_tick)
						core_burnt = core_burnt + (displayer_burnt - reactor.used_cells_last_tick)
					else --new amount would be 0 or below
						reactor.core.get_burnt_result_inventory().clear()
						core_burnt = 0
					end
				end 
				
				--sync displayer to new amount in core
				if core_burnt == 0 then
					reactor.displayer.get_burnt_result_inventory().clear() --sync displayer to new amount in core
					reactor.used_cells_last_tick = 0 --store amount for next tick
				else
					if displayer_burnt == 0 then --safety check
						reactor.displayer.get_burnt_result_inventory().insert(reactor.core.get_burnt_result_inventory()[1]) --sync displayer to new amount in core
					elseif displayer_burnt ~= core_burnt then
						reactor.displayer.get_burnt_result_inventory()[1].count = core_burnt --sync displayer to new amount in core
					end
					reactor.used_cells_last_tick = core_burnt --store amount for next tick
				end
				
			end
		end
		
		-- registered tick for update event of connected reactors list
		if event.tick == global.tick_update_connected_reactors then
			if settings.global["disable-neighbour-check"].value == false then
				logging("---------------------------------------------------------------")
				logging("Updating connected reactors list")
				logging("Tick: " .. event.tick)
				logging("---------------------------------------------------------------")
				for i,reactor in pairs(global.reactors) do
					build_connected_reactors_list(reactor)
				end			
				logging("")
			end	
		end
	
		-- regular update cycle
		if event.tick % TICKS_PER_UPDATE == 0 then
			logging("---------------------------------------------------------------")
			logging("Starting update cycle. Tick: " .. event.tick)
			logging("---------------------------------------------------------------")
			logging("{")
			
			---[[
			
			logging("--------------------------------")
			logging("Debugging global.reactors")
			logging("--------------------------------")
			logging("{")
			for i,r in pairs(global.reactors) do
				logging("Reactor number " .. i)			
				logging("- reactor.id: " .. r.id)
				logging("- reactor.core_id: " .. r.core_id)
				logging("- reactor.core.unit_number: " .. r.core.unit_number)
				logging("- reactor.interface.unit_number: " .. r.interface.unit_number)
				logging("- reactor.eccs.unit_number: " .. r.eccs.unit_number)
				logging("- reactor model: " .. r.core.name)
				logging("- reactor.efficiency: " .. r.efficiency)
				logging("- reactor.state: " .. r.state)
				logging("- reactor.displayer.name: "..r.displayer.name)
				logging("- reactor.power.power_usage: "..r.power.power_usage)
				logging("- reactor.power.energy: "..r.power.energy)
			end
			logging("}")
			--]]		
			
			
			-- every update cycle (default: 15 ticks)
			-- 		calculate temperature changes
			logging("--------------------------------")
			logging("Updating reactor temperature")
			logging("--------------------------------")
			logging("{")
			for i,reactor in pairs(global.reactors) do
				update_reactor_temperature(i,event)
			end			
			logging("}")
			
			
			-- every 4 update cycles (default: 60 ticks = 1 second)
			-- 		check for changed reactor states 
			--		reset cooling tower steam recipe
			if event.tick % (TICKS_PER_UPDATE * 4) == 0 then
			
				logging("--------------------------------")
				logging("Updating reactor states")
				logging("--------------------------------")
				logging("{")
				for i,_ in pairs(global.reactors) do
					update_reactor_states(i,event)
				end
				for i,_ in pairs(global.towers) do
					update_cooling_tower(i)
				end				
				logging("}")
			end
			
			logging("}")
			logging("")
		end -- update event
	end 
	stats_tick(event)
end)

-- registers event for update_connected_reactors list 
function register_update_connected_reactors_event(tick)
	logging("---")
	logging("Registered event update_connected_reactors on tick: " .. tick + settings.global["neighbour-check-delay"].value)
	logging("---")
	global.tick_update_connected_reactors = tick + settings.global["neighbour-check-delay"].value
end

end


-- FUNCTIONS FOR ADDING AND REMOVING BUILDINGS
do

-- functions for all game building/removing events 
do
script.on_event({defines.events.on_built_entity, defines.events.on_robot_built_entity}, function(event)
--script.on_event({}, function(event)

	--game.players[1].print("entity in bp: "..event.created_entity.name)

	--if string.sub(event.created_entity.name,1,17) == "realistic-reactor"
	--	and event.created_entity.name ~= REACTOR_ENTITY_NAME 
	--	and event.created_entity.name ~= BREEDER_ENTITY_NAME 
	--	and event.created_entity.name ~= INTERFACE_ENTITY_NAME
	--	and event.created_entity.name ~= BREEDER_INTERFACE_ENTITY_NAME then
	--	--realistic-reactor-40 - 180 was build by blueprint
	--	-- ToDo: replace it with stopped reactor entity 
	--	
	--	--game.players[1].print("-> Wrong reactor in blueprint")
	--	-- does not trigger on realistic-reactor-118 --> WHY?
	--	
	--end --why is this still here?
	
	
	if event.created_entity.name == REACTOR_ENTITY_NAME or event.created_entity.name == BREEDER_ENTITY_NAME then --build event: reactors

		event.created_entity.active=false
		
		--look for interface
		local interface_entity = event.created_entity.surface.find_entity(INTERFACE_ENTITY_NAME,{event.created_entity.position.x,event.created_entity.position.y+1})
		if not interface_entity then
			interface_entity = event.created_entity.surface.find_entity(BREEDER_INTERFACE_ENTITY_NAME,{event.created_entity.position.x,event.created_entity.position.y+1})
		end
		--look for interface ghost
		local interface_ghost
		if not interface_entity then
			interface_ghost = event.created_entity.surface.find_entities_filtered{ghost_name=INTERFACE_ENTITY_NAME,position={event.created_entity.position.x,event.created_entity.position.y+1}}
			if #interface_ghost == 0 then
				interface_ghost = event.created_entity.surface.find_entities_filtered{ghost_name=BREEDER_INTERFACE_ENTITY_NAME,position={event.created_entity.position.x,event.created_entity.position.y+1}}
			end
		end
		
		if interface_entity == nil and #interface_ghost == 0 then --no interface found
			logging("Building new circuit interface")
			--add new circuit interface
			local interface
			if event.created_entity.name == REACTOR_ENTITY_NAME then --different interfaces for different reactors
				interface = event.created_entity.surface.create_entity
				{
					name = INTERFACE_ENTITY_NAME,
					position = {event.created_entity.position.x, event.created_entity.position.y + 1},
					force = event.created_entity.force
				}
			else
				interface = event.created_entity.surface.create_entity
				{
					name = BREEDER_INTERFACE_ENTITY_NAME,
					position = {event.created_entity.position.x, event.created_entity.position.y + 1},
					force = event.created_entity.force
				}
			end
			add_reactor(event.created_entity,interface)
		elseif interface_entity ~= nil then --built interface found
		
			logging("Found existing circuit interface")
			add_reactor(event.created_entity,interface_entity)
		else --interface ghost found
			event.created_entity.destructible = true
			event.created_entity.minable = true
		end
	elseif event.created_entity.name == INTERFACE_ENTITY_NAME or event.created_entity.name == BREEDER_INTERFACE_ENTITY_NAME then --build event: interfaces
		--look for reactors
		local reactor_entity = event.created_entity.surface.find_entity(REACTOR_ENTITY_NAME,{event.created_entity.position.x,event.created_entity.position.y})
		if not reactor_entity then
			reactor_entity = event.created_entity.surface.find_entity(BREEDER_ENTITY_NAME,{event.created_entity.position.x,event.created_entity.position.y})
		end
		
		if reactor_entity and reactor_entity.position.x == event.created_entity.position.x and reactor_entity.position.y == event.created_entity.position.y-1 then --reactor found and location is correct
			add_reactor(reactor_entity,event.created_entity)
		else
			event.created_entity.destructible = true
			event.created_entity.minable = true
		end
	elseif event.created_entity.name == TOWER_ENTITY_NAME then --build event: cooling tower
		add_cooling_tower(event.created_entity)
	elseif event.created_entity.name == HEAT_PIPE_ENTITY_NAME then
		-- register event for updating list of connected reactors
		register_update_connected_reactors_event(event.tick)
	elseif event.created_entity.name=="constant-combinator" then --build event: constant combinator (manually built)
		--check if there is an interface ghost under the constant combinator
		local interface_ghost = event.created_entity.surface.find_entities_filtered{ghost_name=INTERFACE_ENTITY_NAME,position=event.created_entity.position}
		if #interface_ghost == 0 then
			interface_ghost = event.created_entity.surface.find_entities_filtered{ghost_name=BREEDER_INTERFACE_ENTITY_NAME,position=event.created_entity.position}
		end
		if #interface_ghost > 0 then --interface ghost under constant combinator
			event.created_entity.destroy() --remove constant combinator
			local surface = interface_ghost[1].surface -- store attributes of the interface ghost
			local position = interface_ghost[1].position 
			local force = interface_ghost[1].force
			local name = interface_ghost[1].ghost_name
			interface_ghost[1].revive() --revive interface ghost (maintains circuit wire connections)
			if name == INTERFACE_ENTITY_NAME then --build normal reactor ghost
				surface.create_entity{
					name = "entity-ghost",
					position = {position.x, position.y - 1},
					force = force,
					inner_name = REACTOR_ENTITY_NAME,
					expires = false,
				}
			else --build breeder reactor ghost
				surface.create_entity{
					name = "entity-ghost",
					position = {position.x, position.y - 1},
					force = force,
					inner_name = BREEDER_ENTITY_NAME,
					expires = false,
				}
			end
		end
	end
end)

script.on_event(defines.events.on_pre_ghost_deconstructed, function(event)
if event.ghost.ghost_name == "realistic-reactor-interface" then --interface ghost has been mined via deconstruction planner
	-- look for a reactor
	local reactor_entity = event.ghost.surface.find_entity(REACTOR_ENTITY_NAME,{event.ghost.position.x,event.ghost.position.y})
	local breeder_entity = event.ghost.surface.find_entity(BREEDER_ENTITY_NAME,{event.ghost.position.x,event.ghost.position.y})
	if reactor_entity then --reactor found, recreate interface for free and register the reactor
		local interface = event.ghost.surface.create_entity
			{
				name = INTERFACE_ENTITY_NAME,
				position = event.ghost.position,
				force = event.ghost.force
			}	
		add_reactor(reactor_entity,interface)
	elseif breeder_entity then
		local interface = event.ghost.surface.create_entity
		{
			name = BREEDER_INTERFACE_ENTITY_NAME,
			position = event.ghost.position,
			force = event.ghost.force
		}
		add_reactor(breeder_entity,interface)
	end
end

end)

script.on_event({defines.events.on_pre_player_mined_item, defines.events.on_robot_pre_mined, defines.events.on_entity_died}, function(event)
	if event.entity.name=="entity-ghost" and event.entity.ghost_name == "realistic-reactor-interface" then --interface ghost has been mined manually
		-- look for a reactor
		local reactor_entity = event.entity.surface.find_entity(REACTOR_ENTITY_NAME,{event.entity.position.x,event.entity.position.y})
		local breeder_entity = event.entity.surface.find_entity(BREEDER_ENTITY_NAME,{event.entity.position.x,event.entity.position.y})
		if reactor_entity then --reactor found, recreate interface for free and register the reactor
			local interface = event.entity.surface.create_entity
				{
					name = INTERFACE_ENTITY_NAME,
					position = event.entity.position,
					force = event.entity.force
				}	
			add_reactor(reactor_entity,interface)
		elseif breeder_entity then
			local interface = event.entity.surface.create_entity
			{
				name = BREEDER_INTERFACE_ENTITY_NAME,
				position = event.entity.position,
				force = event.entity.force
			}
			add_reactor(breeder_entity,interface)
		end
	elseif string.sub(event.entity.name,1,17) == "realistic-reactor" then
		remove_reactor(event)
	elseif event.entity.name == TOWER_ENTITY_NAME then
		remove_cooling_tower(event.entity)
	elseif event.entity.name == HEAT_PIPE_ENTITY_NAME then
		-- register event for updating list of connected reactors
		register_update_connected_reactors_event(event.tick)	
	end
end)

end

-- adds core, eccs, lamp, light, electric interface to the reactor core when it is build
function add_reactor(displayer,interface)
	reactor_core = displayer.surface.create_entity{name = "realistic-reactor-1", position = {displayer.position.x, displayer.position.y}, force = displayer.force.name}
	--reactor is actually the reactor core
	reactor_core.destructible = false
	logging("---------------------------------------------------------------")
    logging("Adding new reactor: "..displayer.name)
	logging("Reactor core ID: " .. reactor_core.unit_number)
	
	--interface.operable = false 
	interface.destructible = false 
	interface.minable = false
	
	logging("Reactor ID: " .. interface.unit_number)
	
	--add the eccs
	local eccs = reactor_core.surface.create_entity
	{
		name = BOILER_ENTITY_NAME,
		position = reactor_core.position,
		force = reactor_core.force
	}
	eccs.operable = false 
	eccs.destructible = false
	eccs.minable = false
	
	--add power entity
	local power_name
	if displayer.name == "realistic-reactor-breeder" then
		power_name = BREEDER_POWER_NAME
	else
		power_name = REACTOR_POWER_NAME
	end
	local power = reactor_core.surface.create_entity
	{
		name = power_name,
		position = reactor_core.position,
		force = reactor_core.force
	}
	power.destructible=false
	
	--add light
	local reactor_lamp
	local reactor_light
	if displayer.name == "realistic-reactor-breeder" then
		reactor_lamp = reactor_core.surface.create_entity{name = "rr-black-lamp", position = {reactor_core.position.x+0.62,reactor_core.position.y+0.6}, force = reactor_core.force.name}
		reactor_light = reactor_core.surface.create_entity{name = "rr-black-light", position = {reactor_core.position.x+0.62,reactor_core.position.y+0.6}, force = reactor_core.force.name}
	else
		reactor_lamp = reactor_core.surface.create_entity{name = "rr-black-lamp", position = {reactor_core.position.x-0.62,reactor_core.position.y+0.6}, force = reactor_core.force.name}
		reactor_light = reactor_core.surface.create_entity{name = "rr-black-light", position = {reactor_core.position.x-0.62,reactor_core.position.y+0.6}, force = reactor_core.force.name}
	end
	reactor_lamp.destructible = false
	reactor_light.destructible = false
	
	-- reactor is not active when it is build (state=stopped)
	reactor_core.active=false
	
	table.insert(global.reactors, {
		id = interface.unit_number,	-- ID of the reactor (doesn't change)				
		core_id = reactor_core.unit_number, -- ID of the core (changes when core is replaced)
		core = reactor_core, -- core entity
		interface = interface, -- interface entity
		eccs = eccs, -- eccs entity
		power = power, -- power entity
		displayer = displayer, -- displayer entity
		position = reactor_core.position, -- core position = reactor position
		state = "stopped", -- reactor state
		state_active_since = game.tick, -- state begin
		connected_neighbours_IDs = {}, -- list with IDs of connected reactors 
		neighbours = 0, -- number of connected reactors
		control = interface.get_or_create_control_behavior(), -- control behaviour for interface signals
		efficiency = 0, -- reactor efficiency
		bonus_cells = {}, -- list of bonus cell amount for breeder
		cooling_history=0, -- average power loss through cooling in the last TICKS_PER_UPDATE*8 ticks 
		--> currently not used
		lamp = reactor_lamp, --reactor lamp dummy
		light = reactor_light, --light dummy for the lamp
		interface_warning_tick = 0, --last tick an electricity warning was displayed on the interface (for imitating the game's behaviour)
		interface_warning = nil, --electricity warning dummy (interface)
		cooling_warning_tick = 0, --last tick an electricity warning was displayed on the cooling
		cooling_warning = nil, --electricity warning dummy (cooling)
		power_usage = {starting = 0,cooling = 0, interface = 0}, --portions of the power consumption
		cells_last_tick = 0, --fuel cells in the core on last tick
		used_cells_last_tick = 0,
		fuel_last_tick = 0, -- fuel value in the core on last tick
		fluid_amount_last_tick = 0, -- fluid amount in eccs on last tick
		power_output_last_tick = 0, --power output in MW on last tick (only updated in start or running phase)
		
		signals = -- signals for interface
		{
		  parameters =
		  {
			["core_temperature"] = {signal=SIGNAL_CORE_TEMP, count=0, index=1},
			["state_stopped"] = {signal=SIGNAL_STATE_STOPPED, count=1, index=2},
			["state_starting"] = {signal=SIGNAL_STATE_STARTING, count=0, index=3},
			["state_running"] = {signal=SIGNAL_STATE_RUNNING, count=0, index=4},
			["state_scramed"] = {signal=SIGNAL_STATE_SCRAMED, count=0, index=5},
			["coolant-amount"] = {signal=SIGNAL_COOLANT_AMOUNT, count=0, index=6},
			["power-output"] = {signal=SIGNAL_REACTOR_POWER_OUTPUT, count=0, index=7},
			["uranium-fuel-cells"] = {signal=SIGNAL_URANIUM_FUEL_CELLS, count=0, index=8},
			["used-uranium-fuel-cells"] = {signal=SIGNAL_USED_URANIUM_FUEL_CELLS, count=0, index=9},
			["efficiency"] = {signal=SIGNAL_REACTOR_EFFICIENCY, count=0, index=10},
			["cell-bonus"] = {signal=SIGNAL_REACTOR_CELL_BONUS, count=0, index=11},
			["electric-power"] = {signal=SIGNAL_REACTOR_ELECTRIC_POWER, count=0, index=12}
		  }
		}
	})
	
	logging("-> reactor successfully added")
	logging("")
	
	-- register event for updating list of connected reactors
	register_update_connected_reactors_event(game.tick)
	
end

-- removes other reactor parts when its reactor core is removed
function remove_reactor(event)
	logging("---------------------------------------------------------------")
	logging("Removing reactor")
	--logging("Reactor core ID: " .. dead_reactor_core.unit_number)
	
	for i,reactor in pairs(global.reactors) do
		if reactor.core == event.entity or reactor.displayer == event.entity  then
			dead_reactor_core = reactor.core
			logging("Found reactor, ID: " .. reactor.id)
			logging("-> removing circuit interface, eccs and power entity")


			if reactor.state == "stopped" then
				-- nothing, reactor is not working
			else
				-- working reactor destroyed, causing meltdown
				logging("-> working reactor destroyed, meltdown caused")
				--do the environment damage
				local surface = reactor.core.surface
				surface.create_entity({
					name="medium-explosion",
					position=dead_reactor_core.position
				})
				
				
				if settings.global["RR-explosions"].value then
					surface.create_entity({
						name="fallout-explosion",
						position=dead_reactor_core.position
					})
				end
				if settings.global["clouds_generation"].value > 0 then
					surface.create_entity({
						name="fallout-cloud",
						position=dead_reactor_core.position
					})
					table.insert(global.fallout,{tick = game.tick, position = dead_reactor_core.position, surface = surface})
				else
					surface.create_entity({
						name="permanent-radiation",
						position={dead_reactor_core.position.x,dead_reactor_core.position.y}
					})
					surface.create_entity({
						name="permanent-radiation",
						position={dead_reactor_core.position.x,dead_reactor_core.position.y}
					})
					surface.create_entity({
						name="permanent-radiation",
						position={dead_reactor_core.position.x-4,dead_reactor_core.position.y-4}
					})
					surface.create_entity({
						name="permanent-radiation",
						position={dead_reactor_core.position.x-4,dead_reactor_core.position.y+4}
					})
					surface.create_entity({
						name="permanent-radiation",
						position={dead_reactor_core.position.x+4,dead_reactor_core.position.y-4}
					})
					surface.create_entity({
						name="permanent-radiation",
						position={dead_reactor_core.position.x+4,dead_reactor_core.position.y+4}
					})
				end
			end
			--if reactor.displayer.valid then reactor.displayer.destroy() end
			reactor.eccs.destroy() -- remove eccs
			reactor.power.destroy() -- remove power entity
			--reactor.interface.destroy() -- remove interface entity
			if event.name == defines.events.on_entity_died then
				reactor.interface.destructible=true
				reactor.interface.damage(1000,game.forces.neutral) --retain wire connections in the ghost
			else
				reactor.interface.destroy() -- remove interface entity
			end
			if reactor.core.valid then reactor.core.destroy() end
			if reactor.lamp.valid then reactor.lamp.destroy() end
			if reactor.light.valid then reactor.light.destroy() end
			if reactor.interface_warning and reactor.interface_warning.valid then reactor.interface_warning.destroy() end
			if reactor.cooling_warning and reactor.cooling_warning.valid then reactor.cooling_warning.destroy() end
			
			logging("-> removing global table entry")
			--table.remove(global.reactors, i) -- remove table entry so we stop updating this reactor (edit: this method didnt work)
			global.reactors[i] = nil
		end
	end 
  
	logging("--> reactor successfully removed")
	logging("")
	
	-- register event for updating list of connected reactors
	register_update_connected_reactors_event(event.tick)
	
end

-- adds the steam entity to the cooling tower when it is build
function add_cooling_tower(tower)
-- the steam entity makes happy clouds when the tower is active
-- this is needed because the cooling tower is an electric furnace, and only burner furnaces can produce smoke
  logging("---------------------------------------------------------------")
  logging("Adding new cooling tower with ID: " .. tower.unit_number)
  local steam = tower.surface.create_entity
  {
    name = STEAM_ENTITY_NAME,
    position = tower.position,
    force = tower.force
  }
  steam.operable = false -- disable opening the happy cloud maker's GUI
  steam.destructible = false -- it can't be destroyed (we remove it when the cooling tower dies)
  steam.get_fuel_inventory().insert({name="solid-fuel", count=50}) -- at 1 watt, this is enough fuel to run for 39 years, should suffice
  steam.fluidbox[1] = {name="water", amount=1} -- water for dummy steam puff recipe
  steam.active = false -- start inactive
  table.insert(global.towers,
  {
    id = tower.unit_number,
    entity = tower,
    steam = steam
  })
  logging("-> tower successfully added")
  logging("")
end

-- removes the steam entity when its cooling tower is removed 
function remove_cooling_tower(dead_tower)
logging("---------------------------------------------------------------")
logging("Removing cooling tower ID: " .. dead_tower.unit_number)
  for i,tower in pairs(global.towers) do
    if tower.id == dead_tower.unit_number then
      logging("-> found tower, removing it and all its parts")
	  tower.steam.destroy() -- remove happy cloud maker
      table.remove(global.towers, i) -- remove table entry so we stop trying to update this tower
    end
  end
logging("-> tower successfully removed")
logging("")
end

end


-- FUNCTIONS FOR REACTOR STATE CONTROL AND REACTOR CORE REPLACEMENT
do

-- general function for reactor checks and updates
function update_reactor_states(index, update_event)
	logging("---")
	-- load the reactor
	local reactor = global.reactors[index]
	logging("Updating reactor ID: " .. reactor.id)
	logging("Reactor core ID: " .. reactor.core_id)
	logging("Reactor type: ".. reactor.displayer.name)
	logging("Reactor model: " .. reactor.core.name)
	logging("Reactor state: " .. reactor.state)
	local running_time = math.ceil((update_event.tick - reactor.state_active_since)/60)
	logging("-> state active for (s): " .. running_time) 
	
		
	-- get control signals
	logging("Checking circuit network signals")
	local signal_start = false
	local signal_scram = false
	local green_network = reactor.control.get_circuit_network(defines.wire_type.green)
	if green_network then
		logging("-> Found green circuit network. Network ID: " .. green_network.network_id)
		if green_network.get_signal(SIGNAL_CONTROL_START) > 0 then
			signal_start = true
			logging("--> found green START signal")
		end
		if green_network.get_signal(SIGNAL_CONTROL_SCRAM) > 0 then
			signal_scram = true
			logging("--> found green SCRAM signal")
		end		
	end
	local red_network = reactor.control.get_circuit_network(defines.wire_type.red)
	if red_network then
		logging("-> Found red circuit network. Network ID: " .. red_network.network_id)
		if red_network.get_signal(SIGNAL_CONTROL_START) > 0 then
			signal_start = true
			logging("--> found red START signal")
		end
		if red_network.get_signal(SIGNAL_CONTROL_SCRAM) > 0 then
			signal_scram = true
			logging("--> found red SCRAM signal")
		end		
	end	
	
	-- check for changed states
	logging("Checking for changed reactor state")
	local state_changed = false
	if reactor.state == "stopped" then
		
		if reactor.core.get_fuel_inventory().is_empty() == false
		  and signal_start == true
		  and reactor.power.energy >= (POWER_USAGE_STARTING * TICKS_PER_UPDATE * 4 / 60)
		  and signal_scram == false then
			state_changed = true
			change_reactor_state("starting", reactor, update_event)
		end	
		
	elseif reactor.state == "starting" then
		
		if signal_scram == true then
			state_changed = true
			change_reactor_state("scramed", reactor, update_event)
		elseif running_time >= settings.global["reactor-starting-duration"].value then
			state_changed = true
			change_reactor_state("running", reactor, update_event)
		elseif reactor.power.energy < (POWER_USAGE_STARTING * TICKS_PER_UPDATE * 4 / 60) then
			--ToDo: show alarm on map?
			game.players[1].print("The starting phase of a nuclear reactor was aborted, because the reactor core didn't get enough electric energy from the grid. The fuel cell is lost.")
			state_changed = true
			change_reactor_state("stopped", reactor, update_event)
		end
		
	elseif reactor.state == "running" then
		
		if signal_scram == true then
			state_changed = true
			change_reactor_state("scramed", reactor, update_event)
		elseif reactor.core.get_fuel_inventory().is_empty() == true and reactor.core.burner.heat == 0 then
			
			state_changed = true
			change_reactor_state("stopped", reactor, update_event)
			
		end		
	
	elseif reactor.state == "scramed" then
		
		if running_time >= settings.global["reactor-scram-duration"].value then
			state_changed = true
			change_reactor_state("stopped", reactor, update_event)			
		elseif reactor.core.get_fuel_inventory().is_empty() == true and reactor.core.burner.heat == 0 then
			--ToDo: Problem: scram state ist jetzt active, d.h. reactor holt sich neue cell wenn burner leer ist
			state_changed = true
			change_reactor_state("stopped", reactor, update_event)
		elseif settings.global["choose-a-bug"].value == "stop abrupt" and reactor.core.burner.remaining_burning_fuel < (reactor.signals.parameters["power-output"].count/60*1000000)*(TICKS_PER_UPDATE*5)*(100/reactor.signals.parameters["efficiency"].count) then --tickperupdate *4 but lets be safe..
			local usedfuel = reactor.core.burner.currently_burning.burnt_result.name
			state_changed = true
			change_reactor_state("stopped", reactor, update_event)
			reactor.core.get_burnt_result_inventory().insert({name = usedfuel, count = 1})
		end
		
	end	
	running_time = math.ceil((update_event.tick - reactor.state_active_since)/60)
	
	if state_changed == false then
		logging("-> reactor state unchanged")
	end
	
	--update reactor signals
	if reactor.state  == "starting" then
		reactor.signals.parameters["state_starting"].count = settings.global["reactor-starting-duration"].value - running_time
	end
	if reactor.state  == "scramed" then
		reactor.signals.parameters["state_scramed"].count = settings.global["reactor-scram-duration"].value - running_time
	end	
	if reactor.state  == "stopped" then
		-- reactor is not running
		reactor.signals.parameters["power-output"].count = 0
		reactor.signals.parameters["efficiency"].count = 0
		reactor.signals.parameters["cell-bonus"].count = 0
	end
	

	-- check reactor and replace it with the appropriate version depending on temperature and connected reactors
	if reactor.state == "running" or reactor.state == "starting" or reactor.state == "scramed" then
		
		logging("---Updating reactor stats---")
		
		-- check how many running reactors are connected
		local reactor_neighbours = 1
		logging("Checking running reactor neighbours:")
		local neighbours = 0
		for i,rid in pairs(reactor.connected_neighbours_IDs) do
			--logging("- connected reactor ID: " .. rid)
			neighbours = neighbours + 1
		end
		logging("- Number of connected reactors: " .. neighbours)		
		
		for i,id in pairs(reactor.connected_neighbours_IDs) do
			logging("- checking connected reactor ID: " .. id)
			
			if id == reactor.id then
				-- same reactor, do nothing
				logging("-> this is the current reactor")					
			else
				-- found another reactor, check if it is running
				for i,global_reactor in pairs(global.reactors) do
					if global_reactor.id == id then
						logging("-> loaded connected reactor ID: " .. global_reactor.id)
						if global_reactor.state == "running" then
							reactor_neighbours = reactor_neighbours +1
							logging("--> reactor is running, bonus: " .. reactor_neighbours)
						else
							logging("--> reactor is not running, no bonus")
						end
					end
				end
			end
		end
		logging("-> Running reactor neighbours: " .. reactor_neighbours)	
		reactor.neighbours = reactor_neighbours

		
		--load reactor parameters: 
		-- power , efficiency, bonus_cells, (max_power, max_efficiency)
		logging("Calculating reactor stats:")
		local reactor_parameters
		if settings.global["calculate-stats-function"].value == "ingo's formulas" then
			reactor_parameters = calculate_stats_ingo(reactor,running_time)
		else
			reactor_parameters = calculate_stats_ownly(reactor,running_time)
		end
		-- -> in stats function
		-- if reactor.state  == "starting" then
			-- reactor.signals.parameters["state_starting"].count = settings.global["reactor-starting-duration"].value - running_time + 1 --ToDo: starting phase should end, when first fuel cell is half empty
			-- reactor_parameters.power = math.floor(reactor_parameters.power * ((running_time)/settings.global["reactor-starting-duration"].value)^2)
			-- --reactor_parameters.efficiency = reactor_parameters.efficiency
			-- reactor_parameters.bonus_cells = 0
		-- end
		-- if reactor.state  == "scramed" then
			-- reactor.signals.parameters["state_scramed"].count = settings.global["reactor-scram-duration"].value - running_time + 1
			-- reactor_parameters.power = math.floor(reactor_parameters.power * ((settings.global["reactor-scram-duration"].value - (running_time)/3.5)/settings.global["reactor-scram-duration"].value)^11+0.45)
			-- --reactor_parameters.efficiency = reactor_parameters.efficiency
			-- reactor_parameters.bonus_cells = 0
		-- end	
			
		logging("-> Temperature="..reactor.core.temperature.." PowerOutput="..reactor_parameters.power.." Efficiency="..math.floor(reactor_parameters.efficiency).." BonusCellAmount: "..reactor_parameters.bonus_cells)
		
		--apply material bonus by adding empty fuel cell
		local burnt_result 
		logging("Applying empty fuel cell bonus:")
		if reactor.core.burner.currently_burning == nil then
			-- burner is empty, can't read fuel type
			-- cell bonus is skipped for now
			-- ToDo: save it in global.reactors and add it during the next update
		else
			logging("- currently burning: "..reactor.core.burner.currently_burning.name)
			if reactor_parameters.bonus_cells == 0 then
				-- do nothing
				-- normal reactor or breeder too cold
				logging("-> adding nothing, too cold or no breeder reactor")
			else
				-- add bonus to current empty fuel cell amount
				
				-- burn duration in ticks
				local burn_duration = ((reactor.core.burner.currently_burning.fuel_value/1000000)*(reactor_parameters.efficiency/100))/reactor_parameters.power*15/TICKS_PER_UPDATE 
				logging("- burn cycle duration (s): "..burn_duration)
				burnt_result = reactor.core.burner.currently_burning.burnt_result
				--game.players[1].print(reactor.bonus_cells[burnt_result.name])
				reactor.bonus_cells[burnt_result.name] = (reactor.bonus_cells[burnt_result.name] or 0) + reactor_parameters.bonus_cells / burn_duration
			end
		end
		
		
		if burnt_result and reactor.bonus_cells[burnt_result.name] and reactor.bonus_cells[burnt_result.name] > 1 then
			-- add used-up-uranium-fuel-cell
			logging("Inserting used-up-uranium-fuel-cell:")
			--game.players[1].print("bonuscell")
			if reactor.displayer.get_burnt_result_inventory().is_empty() == false then
				--burnt fuel inventory not empty
				local current_empty_cell = reactor.displayer.get_inventory(defines.inventory.burnt_result)
				if current_empty_cell[1].name == burnt_result.name and current_empty_cell[1].count < burnt_result.stack_size then
					-- same cell
					reactor.displayer.get_burnt_result_inventory().insert({name=burnt_result.name, count=1})
					reactor.bonus_cells[burnt_result.name] = reactor.bonus_cells[burnt_result.name] -1
					logging("-> inserted")
				else
					--different cell, wait
					logging("-> can't insert, different used up cell present. waiting...")
				end
			else
				--burnt fuel inventory empty
				reactor.displayer.get_burnt_result_inventory().insert({name=burnt_result.name,count=1})
				reactor.bonus_cells[burnt_result.name] = reactor.bonus_cells[burnt_result.name] -1
				logging("-> inserted")
			end	
			
		end
		
		
		-- update reactor signals 
		reactor.signals.parameters["power-output"].count = reactor_parameters.power
		reactor.signals.parameters["efficiency"].count = math.floor(reactor_parameters.efficiency)
		reactor.signals.parameters["cell-bonus"].count = reactor_parameters.bonus_cells*100
		reactor.efficiency = reactor_parameters.efficiency
		if reactor.state == "running" or reactor.state == "starting" then
			reactor.power_output_last_tick = reactor_parameters.power
		end
		
		
		-- replace reactor with updated level version
		logging("Replace reactor model:")
		local reactor_to_build = "realistic-reactor-"..math.max(1,reactor_parameters.power)
		logging("-Reactor to be: "..reactor_to_build)
		logging("-Current reactor: "..reactor.core.name)
		if reactor.core.name==reactor_to_build then
			logging("-> Reactor already build.")
		elseif not (reactor.state =="scramed" and settings.global["choose-a-bug"].value == "free energy in scram") then
			replace_reactor(reactor,reactor_to_build)
		end
		
	
	end
	
	
	-- UPDATE DISPLAYER --
	reactor.displayer.temperature = reactor.core.temperature
	
end

-- changes reactor state
function change_reactor_state(new_state, reactor, update_event)
	logging("-> changing reactor state to: " .. new_state)
	reactor.state = new_state
	reactor.state_active_since = update_event.tick
	reactor.lamp.destroy()
	reactor.light.destroy()
	local light_color = "black"
	if new_state == "stopped" then
		-- set signals
		reactor.signals.parameters["state_stopped"].count = 1
		reactor.signals.parameters["state_starting"].count = 0
		reactor.signals.parameters["state_running"].count = 0
		reactor.signals.parameters["state_scramed"].count = 0
		-- configure reactor
		replace_reactor(reactor,"realistic-reactor-1")
		reactor.displayer.active = false
		reactor.core.active = false
		reactor.displayer.minable = true
	elseif new_state == "starting" then
		--set signals
		reactor.signals.parameters["state_stopped"].count = 0
		reactor.signals.parameters["state_starting"].count = 1
		reactor.signals.parameters["state_running"].count = 0
		reactor.signals.parameters["state_scramed"].count = 0	
		-- configure reactor
		reactor.displayer.active = true
		reactor.core.active = true
		--reactor.debug_start_cell = true
		reactor.displayer.minable = false
		light_color = "yellow"
	elseif new_state == "running" then	
		-- set signals
		reactor.signals.parameters["state_stopped"].count = 0
		reactor.signals.parameters["state_starting"].count = 0
		reactor.signals.parameters["state_running"].count = 1
		reactor.signals.parameters["state_scramed"].count = 0
		-- configure reactor
		reactor.displayer.active = true
		reactor.core.active = true
		
		reactor.displayer.minable = false
		light_color = "green"
	elseif new_state == "scramed" then
		--set signals
		reactor.signals.parameters["state_stopped"].count = 0
		reactor.signals.parameters["state_starting"].count = 0
		reactor.signals.parameters["state_running"].count = 0
		reactor.signals.parameters["state_scramed"].count = 1	
		-- configure reactor
		if settings.global["choose-a-bug"].value == "free energy in scram" then
			reactor.core.active = false
			reactor.displayer.active = false
			if reactor.core.burner.currently_burning then
				reactor.core.get_burnt_result_inventory().insert({name = reactor.core.burner.currently_burning.burnt_result.name, count = 1})
			end
			reactor.core.burner.currently_burning = nil
			reactor.core.burner.remaining_burning_fuel = 0
		else
			reactor.core.active = true
			reactor.displayer.active = true
		end
		reactor.displayer.minable = false
		light_color = "red"
	end
	if reactor.displayer.name == "realistic-reactor-breeder" then
		reactor.lamp = reactor.core.surface.create_entity{name = "rr-"..light_color.."-lamp", position = {reactor.core.position.x+0.62,reactor.core.position.y+0.6}, force = reactor.core.force.name}
		reactor.light = reactor.core.surface.create_entity{name = "rr-"..light_color.."-light", position = {reactor.core.position.x+0.62,reactor.core.position.y+0.6}, force = reactor.core.force.name}
	else
		reactor.lamp = reactor.core.surface.create_entity{name = "rr-"..light_color.."-lamp", position = {reactor.core.position.x-0.62,reactor.core.position.y+0.6}, force = reactor.core.force.name}
		reactor.light = reactor.core.surface.create_entity{name = "rr-"..light_color.."-light", position = {reactor.core.position.x-0.62,reactor.core.position.y+0.6}, force = reactor.core.force.name}
	end
	reactor.lamp.destructible = false
	reactor.light.destructible = false
end

-- replaces the reactor core entity with another one
function replace_reactor(reactor, new_reactor_entity_name)
	local temp = reactor.core.temperature
	logging("Building reactor model: "..new_reactor_entity_name)
		
	-- get table entry index for reactor to be replaced
	local table_index
	for i,table_reactor in pairs(global.reactors) do
		if table_reactor.id == reactor.id then
			table_index = i
		end
	end
	logging("- old reactor ID: " .. reactor.id)
	logging("- old reactor core ID: " .. reactor.core_id)
	
	-- create new reactor core
	new_reactor_core = reactor.core.surface.create_entity
	{
		name = new_reactor_entity_name,
		position = reactor.core.position,
		force = reactor.core.force		
	}
	logging("- new reactor core ID: " .. new_reactor_core.unit_number)
	
	-- copy everything from old core to new core 
	new_reactor_core.copy_settings(reactor.core) --(what is this actually copying???)
	new_reactor_core.temperature = temp
	logging("-> updated temperature: " .. new_reactor_core.temperature)
	-- transfer burner heat and remaining fuel in burner
	if reactor.state == "stopped" and not (settings.global["choose-a-bug"].value == "stop half-empty") then
		-- do nothing (don't transfer burner heat to a stopped or scramed reactor)
		if settings.global["choose-a-bug"].value == "consume additional cell" and reactor.core.burner.currently_burning then
			reactor.core.get_burnt_result_inventory().insert({name = reactor.core.burner.currently_burning.burnt_result.name, count = 1})
		end
		logging("-> burner settings not transferred, state: stopped or scramed")
	else
		-- transfer current burner settings
		if reactor.core.burner.heat > 0 then
			--new_reactor_core.burner.currently_burning = game.item_prototypes["uranium-fuel-cell"]
			new_reactor_core.burner.currently_burning = reactor.core.burner.currently_burning
			new_reactor_core.burner.heat = reactor.core.burner.heat
			logging("-> updated burner heat: " .. new_reactor_core.burner.heat)
			new_reactor_core.burner.remaining_burning_fuel = reactor.core.burner.remaining_burning_fuel
			logging("-> updated burner remaining_burning_fuel: " .. new_reactor_core.burner.remaining_burning_fuel)
		else
			logging("-> burner settings not transferred, empty")
		end			
	end
	
	new_reactor_core.minable = false -- core should be unminable in any case cause its removed via script
	-- set minable and active
	--if reactor.state == "running" then
	--	new_reactor_core.active = true	-- not necessary, done in change_reactor_state function
	--end

	-- transfer fuel cells and empty fuel cells
	if reactor.core.get_fuel_inventory().is_empty() == false then
		-- get current fuel cell
		local fuel = reactor.core.get_inventory(defines.inventory.fuel)
		logging("-> fuel cell name: " .. fuel[1].name)
		logging("-> moved fuel cells, count: " .. new_reactor_core.get_fuel_inventory().insert(fuel[1]))
	else
		logging("-> no fuel cells moved")
	end
	if not reactor.core.get_burnt_result_inventory().is_empty() then
		new_reactor_core.get_burnt_result_inventory().insert(reactor.core.get_burnt_result_inventory()[1])
	end
	--if reactor.core.get_burnt_result_inventory().is_empty() == false then
	--	local usedfuel = reactor.core.get_inventory(defines.inventory.burnt_result)
	--	logging("-> used fuel cell name: " .. usedfuel[1].name)
	--	logging("-> moved used fuel cells, count: " .. new_reactor_core.get_burnt_result_inventory().insert(usedfuel[1]))
	--else
	--	logging("-> no used fuel cells moved")
	--end	
	

	-- destroy old core
	reactor.core.destroy()
	-- store new core
	reactor.core = new_reactor_core
	--update reactor core id with new core unit_number
	reactor.core_id = new_reactor_core.unit_number
	-- update table entry
	global.reactors[table_index]=reactor
	
	logging("-> reactor replaced")
	logging("- new reactor ID: " .. reactor.id)
	logging("- new reactor core ID: " .. reactor.core_id)
end

end


-- FUNCTIONS FOR REACTOR TEMPERATURE CHANGES
do

-- updates reactor core temperature
function update_reactor_temperature(index, update_event)
	logging("---")

	-- load the reactor
	local reactor = global.reactors[index]
	logging("Updating reactor ID: " .. reactor.id)
	logging("Reactor core ID: " .. reactor.core_id)
	logging("Reactor core temperature: " .. reactor.core.temperature)
	
	--apply efficiency by adding a fuel bonus
	logging("Applying fuel bonus:")
	if reactor.efficiency > 0 then
		local fuel_bonus = (reactor.signals.parameters["power-output"].count/60*1000000)*(TICKS_PER_UPDATE)*(100/reactor.efficiency-1)
		logging("fuel_bonus: "..fuel_bonus)
		if reactor.core.burner.remaining_burning_fuel - fuel_bonus > 0 then
			reactor.core.burner.remaining_burning_fuel = reactor.core.burner.remaining_burning_fuel - fuel_bonus
		else
			reactor.core.burner.remaining_burning_fuel = 0
		end
		logging("-> adding fuel: "..(fuel_bonus*(-1)))
	else
		logging("-> no fuel bonus, efficiency=0 (reactor stopped)")
	end	
	
	--apply starting power consumption
	if reactor.state == "starting" then
		--apply cooling power consumption
		reactor.power_usage.starting = math.floor(POWER_USAGE_STARTING / 60)
	else
		reactor.power_usage.starting = 0
	end
	
	-- do environment cooling if state is stopped
	if reactor.state == "stopped" then
		if reactor.core.temperature > 15 then
			logging("Reactor state is stopped, apply cooling by the environment")
			
			--[[
			local Tmix_env = ((reactor.core.temperature * REACTOR_MASS) + (15 * 10000)) / (REACTOR_MASS + 10000)
			logging("Tmix_env: " .. Tmix_env)
			local Tdelta_reactor_env = reactor.core.temperature - Tmix_env	
			logging("Tdelta_reactor_env: " .. Tdelta_reactor_env)
			local Tchange_reactor_env = (Tdelta_reactor_env * CHANGE_MULTIPLIER * 0.1)
			logging("-> Tchange_reactor_env: " .. Tchange_reactor_env)
			if (reactor.core.temperature - Tchange_reactor_env) > 15 then
				reactor.core.temperature = reactor.core.temperature - Tchange_reactor_env
			else
				reactor.core.temperature = 15
			end
			]]
			
			if reactor.core.temperature > 15.125 then
				reactor.core.temperature = reactor.core.temperature - 0.125
			else
				reactor.core.temperature = 15
			end
						
			logging("Reactor core temperature after environment cooling: " .. reactor.core.temperature)
		end
	end
	
	-- do decay heat effect if state is scramed
	if reactor.state == "scramed" and settings.global["choose-a-bug"].value == "free energy in scram" then
		logging("Reactor state is scramed, apply decay heat effect:")
		reactor.core.temperature = reactor.core.temperature + reactor.signals.parameters["power-output"].count/20
		logging("Reactor core temperature after heat decay effect: " .. reactor.core.temperature)
	end
	reactor.power_usage.cooling = 0 -- supposed to be overwritten
	reactor.cooling_history = reactor.cooling_history*7/8 --devaluing previous history (only used by graph)
	-- apply cooling power consumption
	-- (if eccs fluid amount has changed since last tick, then power consumption is applied)
	if settings.global["static-cooling-power-consumption"].value then
		if reactor.eccs.fluidbox[1] ~= nil then
			if math.abs(reactor.fluid_amount_last_tick - reactor.eccs.fluidbox[1].amount) > 0 then
				--apply cooling power consumption
				reactor.power_usage.cooling = math.floor(POWER_USAGE_COOLING / 60)
				--game.players[1].print("cooling")
			else
				--reactor.power_usage.cooling = 0
				--game.players[1].print("not cooling")
			end
		end
	end
	--remove cooling power alert (gfx)
	if game.tick - reactor.cooling_warning_tick >=60  and reactor.cooling_warning and reactor.cooling_warning.valid then
		reactor.cooling_warning.destroy()
	end
 	-- check fluid level in eccs and do cooling
	logging("Applying ECCS cooling effect")
	if reactor.eccs.fluidbox[1] == nil then
		--do nothing, no coolant
		logging("-> eccs empty")
		
		--reset the coolant fluid signal
		reactor.signals.parameters["coolant-amount"].count = 0		
		
	else
		local fluid = reactor.eccs.fluidbox[1]
		logging("- CHANGE_MULTIPLIER: " .. CHANGE_MULTIPLIER)
		logging("- Reactor mass: " .. REACTOR_MASS)
		logging("- Reactor temperature: " .. reactor.core.temperature)
		logging("- Fluid amount: " .. fluid.amount)
		logging("- Fluid temperature: " .. fluid.temperature)
		if fluid.amount < 100 then
			--do nothing, not enough coolant or fluid temperature at max
			logging("-> not enough fluid in eccs")
			--update the coolant fluid signals
			reactor.signals.parameters["coolant-amount"].count = fluid.amount
						
		elseif reactor.power.energy < (POWER_USAGE_COOLING * TICKS_PER_UPDATE / 60) then
			-- do nothing, not enough electric energy
			--game.players[1].print("not cooling.. energy: "..reactor.power.energy.." / "..(POWER_USAGE_COOLING * TICKS_PER_UPDATE / 60))
			logging("-> not enough electric energy for ECCS")
			--update the coolant fluid signals
			reactor.signals.parameters["coolant-amount"].count = fluid.amount
			--ToDo: show alert on map?
			game.players[1].print("A nuclear reactor can't be cooled by its ECCS, because it doesn't enough electric energy from the grid.")
			if game.tick - reactor.cooling_warning_tick >=120 then
				reactor.cooling_warning=reactor.displayer.surface.create_entity{name="rr-electricity-warning", position = reactor.displayer.position, force = "neutral"}
				reactor.cooling_warning.destructible = false
				reactor.cooling_warning_tick = game.tick
			end
			
		else
			
			--calculate the mixing temperature with richmann's mixing rule
			local Tmix = ((reactor.core.temperature * REACTOR_MASS) + (fluid.amount * fluid.temperature)) / (REACTOR_MASS + fluid.amount)
			logging("Tmix: " .. Tmix)
			
			-- check which is hotter and cool/heat accordingly
			local Tdelta_reactor = reactor.core.temperature - Tmix
			local Tdelta_fluid = fluid.temperature - Tmix
			logging("Tdelta_reactor: " .. Tdelta_reactor)
			logging("Tdelta_fluid: " .. Tdelta_fluid)
			
			if Tdelta_reactor > Tdelta_fluid then
				-- reactor is hotter than fluid (that is how it should be)
				logging("-> cooling reactor, heating fluid")
				
				local Tchange_reactor = (Tdelta_reactor * CHANGE_MULTIPLIER)
				local Tchange_fluid = (Tdelta_fluid * CHANGE_MULTIPLIER)
				logging("-> Tchange_reactor: " .. Tchange_reactor)
				logging("-> Tchange_fluid: " .. Tchange_fluid)
				
				if fluid.temperature - Tchange_fluid > 100 then
					-- resulting fluid would be too hot (max 100°, set by game), reduce cooling amount
					logging("--> fluid too hot, cooling with reduced efficiency")
					
					--reduce both temperature changes by the same factor, so that resulting fluid is = 100°
					efficiency_factor = (100 - fluid.temperature) / (Tdelta_fluid * CHANGE_MULTIPLIER * -1)
					
					logging("--> efficiency_factor: " .. efficiency_factor)
					logging("--> new Tchange_reactor: " .. Tchange_reactor * efficiency_factor)
					logging("--> new Tchange_fluid: " .. Tchange_fluid * efficiency_factor)
					
					--cool the reactor core
					reactor.core.temperature = reactor.core.temperature - (Tchange_reactor * efficiency_factor)
					reactor.cooling_history = reactor.cooling_history + (Tchange_reactor * efficiency_factor * 20 / 8)
					if not settings.global["static-cooling-power-consumption"].value then
						reactor.power_usage.cooling = math.floor(Tchange_reactor * efficiency_factor *POWER_USAGE_COOLING / 60)
						reactor.cooling_history = reactor.cooling_history + reactor.power_usage.cooling*60/1000000/8* settings.global["energy_consumption_multiplier"].value
					end
					--heat the fluid
					fluid.temperature = fluid.temperature - (Tchange_fluid * efficiency_factor)
					

					
					
				else
					logging("--> cooling with maximum efficiency")
					--cool the reactor core
					reactor.core.temperature = reactor.core.temperature - Tchange_reactor
					reactor.cooling_history = reactor.cooling_history + (Tchange_reactor * 20 / 8)
					if not settings.global["static-cooling-power-consumption"].value then
						reactor.power_usage.cooling = math.floor(Tchange_reactor * POWER_USAGE_COOLING / 60)
						reactor.cooling_history = reactor.cooling_history + reactor.power_usage.cooling*60/1000000/8* settings.global["energy_consumption_multiplier"].value
					end
					--heat the fluid
					fluid.temperature = fluid.temperature - Tchange_fluid						
					
					
				end			
				
			else
				-- fluid is hotter than reactor (it's possible to heat the reactor with a hot fluid)
				logging("-> heating reactor, cooling fluid")
				
				local Tchange_reactor = (Tdelta_reactor * CHANGE_MULTIPLIER)
				local Tchange_fluid = (Tdelta_fluid * CHANGE_MULTIPLIER)
				logging("-> Tchange_reactor: " .. Tchange_reactor)
				logging("-> Tchange_fluid: " .. Tchange_fluid)
				
				if reactor.core.temperature - Tchange_reactor > 100 then
					--do nothing, reactor is already 100°
					-- this is necessary, because theoretically it would be possible to heat the reactor with steam to 1000°, thus causing the meltdown
					logging("--> no change, reactor too hot")
				else
					--heat the reactor core
					reactor.cooling_history = reactor.cooling_history + (Tchange_reactor * 20 / 8)
					if not settings.global["static-cooling-power-consumption"].value then
						reactor.power_usage.cooling = math.floor(Tchange_reactor * POWER_USAGE_COOLING / 60)
						reactor.cooling_history = reactor.cooling_history + reactor.power_usage.cooling*60/1000000/8* settings.global["energy_consumption_multiplier"].value
					end
					reactor.core.temperature = reactor.core.temperature - Tchange_reactor
					--cool the fluid
					fluid.temperature = fluid.temperature - Tchange_fluid	


				end
				
			end

			logging("-> Reactor temperature: " .. reactor.core.temperature)
			logging("-> Fluid temperature: " .. fluid.temperature)
			
			--update the coolant fluid signal
			reactor.signals.parameters["coolant-amount"].count = fluid.amount
			
			--save the fluid back to the eccs
			reactor.eccs.fluidbox[1] = fluid
			
		end -- not enough coolant or not enough energy
	
	end --empty eccs

	
	-- get amount of fuel cells and set signal
	if reactor.displayer.get_fuel_inventory().is_empty() == true then
		reactor.signals.parameters["uranium-fuel-cells"].count = 0
		reactor.signals.parameters["uranium-fuel-cells"].signal = {type="item", name="uranium-fuel-cell"}
	else
		reactor.signals.parameters["uranium-fuel-cells"].signal = {type="item", name=reactor.displayer.get_fuel_inventory()[1].name}
		reactor.signals.parameters["uranium-fuel-cells"].count = reactor.displayer.get_fuel_inventory()[1].count
	end

	-- get amount of used fuel cells and set signal
	if reactor.displayer.get_burnt_result_inventory().is_empty() == true then
		reactor.signals.parameters["used-uranium-fuel-cells"].count = 0
		reactor.signals.parameters["used-uranium-fuel-cells"].signal = {type="item", name="used-up-uranium-fuel-cell"}
	else
		reactor.signals.parameters["used-uranium-fuel-cells"].signal = {type="item", name=reactor.displayer.get_burnt_result_inventory()[1].name}
		reactor.signals.parameters["used-uranium-fuel-cells"].count = reactor.displayer.get_burnt_result_inventory()[1].count
	end		
	
	-- destroy energy warning entity
	if game.tick - reactor.interface_warning_tick >=60  and reactor.interface_warning and reactor.interface_warning.valid then
		reactor.interface_warning.destroy()
	end
	-- update reactor signals
	reactor.signals.parameters["electric-power"].count = (reactor.power.energy/17000000)*100
	reactor.signals.parameters["core_temperature"].count = reactor.core.temperature
	
	-- check if interface has enough power
	if reactor.power.energy > (POWER_USAGE_INTERFACE * TICKS_PER_UPDATE / 60) then   --- TICKS_PER_UPDATE
		-- show updated signals on interface
		reactor.control.parameters = reactor.signals
		-- apply power consumption for interface
		reactor.power_usage.interface = math.floor(POWER_USAGE_INTERFACE/60)
	else
		reactor.power_usage.interface = 0
		-- disable interface signals
		reactor.control.parameters = {parameters={}}
		-- show energy warning entity
		if game.tick - reactor.interface_warning_tick >=120 then
			reactor.interface_warning=reactor.interface.surface.create_entity{name="rr-electricity-warning", position = reactor.interface.position, force = "neutral"}
			reactor.interface_warning.destructible = false
			reactor.interface_warning_tick = game.tick
		end
	end
	
	-- set power usage value of power entity
	logging("Reactor power consumption:")
	logging("-Starting: "..reactor.power_usage.starting)
	logging("-Cooling: "..reactor.power_usage.cooling)
	logging("-Interface: "..reactor.power_usage.interface)
	-- if reactor and reactor.power and reactor.power.valid then
	reactor.power.power_usage = (reactor.power_usage.starting + reactor.power_usage.cooling + reactor.power_usage.interface) * settings.global["energy_consumption_multiplier"].value
	-- end
	logging("-> "..reactor.power.power_usage)
		
	
	

	
	--save fluid amount (settings.global["static-cooling-power-consumption"].value)
	if reactor.eccs.fluidbox[1] ~= nil then
		reactor.fluid_amount_last_tick = reactor.eccs.fluidbox[1].amount
	end	
	
	-- start nuclear meltdown
	if reactor.core.temperature > 998 then
		logging("Core temperature > 998°, MELTDOWN")
		-- destroy the reactor core (will trigger meltdown)
		reactor.displayer.damage(1000,game.forces.neutral)		
	end

end

end


-- FUNCTIONS FOR BUILDING THE LIST OF CONNECTED REACTORS
do

-- updates reactor.connected_neighbours_IDs with the IDs of the current connected reactors
function build_connected_reactors_list(reactor)
	-- find reactor neighbours
	logging("Reactor.id: " .. reactor.id)
	logging("Reactor.core_id: " .. reactor.core_id)
	logging("Reactor position. X: " .. reactor.position.x .. " Y: " .. reactor.position.y)
	local surface = reactor.core.surface
	local hp_neighbour_entities_ew
	local hp_neighbour_entities_n
	local table_of_heat_pipes_to_check
	
	-- reset global table
	global.connected_reactors = {}	
	global.all_heat_pipes = {}
	-- reset connected_neighbours_IDs and insert own id
	-- (a reactor is always connected with itself)
	reactor.connected_neighbours_IDs = {}
	table.insert(reactor.connected_neighbours_IDs,reactor.id)	
	
	-- load all heat pipes
	global.all_heat_pipes = surface.find_entities_filtered{name='heat-pipe'}
	-- table empty check
	if next(global.all_heat_pipes) == nil then
		logging("-> no heat pipes on map, skipping check")
		--goto end_function
	else
	
		-- find all heat pipes next to reactor
		hp_neighbour_entities_ew = surface.find_entities_filtered{area = {{reactor.position.x-2,reactor.position.y-1},{reactor.position.x+2,reactor.position.y}}, name='heat-pipe'} --east and west
		hp_neighbour_entities_n = surface.find_entities_filtered{area = {{reactor.position.x-1,reactor.position.y-2},{reactor.position.x+1,reactor.position.y-2}}, name='heat-pipe'} -- north
		table_of_heat_pipes_to_check = union_tables(hp_neighbour_entities_ew,hp_neighbour_entities_n)
		
		-- loop through all 7 start heat pipes 
		logging("Checking connected heat pipes")
		if next(table_of_heat_pipes_to_check) == nil then
			logging("-> no heat pipes connected")
		else
			for i,hp in ipairs(table_of_heat_pipes_to_check) do
				logging("- checking heat pipe, ID: " .. hp.unit_number .. " X:" .. hp.position.x .. " Y:" .. hp.position.y)
				
				-- load connected reactors
				local connected_reactors = get_connected_reactors(hp)
				for i,connected_reactor in pairs(connected_reactors) do
					-- check if connected reactor is already in global list
					local is_in_list = false
					for i,list_reactor in pairs(global.connected_reactors) do
						if connected_reactor.id == list_reactor.id then
							-- reactor is already in list
							is_in_list = true
						end
					end
					-- if not add it to list
					if is_in_list == false then
						table.insert(global.connected_reactors,connected_reactor)
						logging("--> found NEW connected reactor, ID: " .. connected_reactor.id)
					end
				end			
				
				-- load connected heat pipes
				local connected_heat_pipes = get_connected_heat_pipes(hp)
				for i,connected_heat_pipe in pairs(connected_heat_pipes) do
					-- check if heat pipe is already in list
					local is_in_list2 = false
					for i,list_heat_pipe in pairs(table_of_heat_pipes_to_check) do
						if connected_heat_pipe.unit_number == list_heat_pipe.unit_number then
							is_in_list2 = true
						end
					end
					-- if not add it to list
					if is_in_list2 == false then
						table.insert(table_of_heat_pipes_to_check,connected_heat_pipe)
						--logging("--> found NEW connected heat pipe, ID: " .. connected_heat_pipe.unit_number)
					end
				end
				
						
			end -- for
			
			-- add connected reactors to list
			reactor.connected_neighbours_IDs = {}
			for i,connected_reactor in pairs(global.connected_reactors) do
				table.insert(reactor.connected_neighbours_IDs,connected_reactor.id)
			end		
			
		end -- no heat pipes connected to reactor
	end -- no heat pipes on map
	
	
	logging("Connected reactors (including self)")
	for k,rid in pairs(reactor.connected_neighbours_IDs) do
		logging("- ID: " .. rid)
	end		
	logging("---")
end

-- returns the connected reactors as array or nil
function get_connected_reactors(heat_pipe)
	local hp = heat_pipe
	local result = {}
	
	for i,reactor in pairs(global.reactors) do
		-- pipes on west reactor side
		if reactor.position.x - 2 == hp.position.x then
			if reactor.position.y == hp.position.y or reactor.position.y - 1 == hp.position.y then
				--logging("--> connected to reactor, ID: " .. reactor.id)
				table.insert(result, reactor)
			end
		end
		-- pipes on east reactor side
		if reactor.position.x + 2 == hp.position.x then
			if reactor.position.y == hp.position.y or reactor.position.y - 1 == hp.position.y then
				--logging("--> connected to reactor, ID: " .. reactor.id)
				table.insert(result, reactor)
			end		
		end		
		-- pipes on north reactor side
		if reactor.position.y - 2 == hp.position.y then
			if reactor.position.x - 1 == hp.position.x
			   or reactor.position.x == hp.position.x
			   or reactor.position.x + 1 == hp.position.x then
					--logging("--> connected to reactor, ID: " .. reactor.id)
					table.insert(result, reactor)
			end
		end		
	end -- for
	
	return result
end

-- returns the connected heat pipes as array or nil
function get_connected_heat_pipes(heat_pipe)
	local hp = heat_pipe
	local result = {}	
	
	for i,hplist in pairs(global.all_heat_pipes) do
		
		if hp.position.x == hplist.position.x then
			if hp.position.y == hplist.position.y + 1 or hp.position.y == hplist.position.y - 1 then
				--logging("--> connected to heat pipe, ID: " .. hplist.unit_number)
				table.insert(result, hplist)
			end
		end
		
		if hp.position.y == hplist.position.y then
			if hp.position.x == hplist.position.x + 1 or hp.position.x == hplist.position.x - 1 then
				--logging("--> connected to heat pipe, ID: " .. hplist.unit_number)
				table.insert(result, hplist)
			end
		end		
		
	end
	
	
	return result
end

end


-- FUNCTIONS FOR COOLING TOWER
do

-- resets the cooling tower steam producing crafting process
function update_cooling_tower(index)
  local tower = global.towers[index]
  if tower and tower.entity.valid then
    -- only show steam puffs if cooling tower is actively working and not backed up
    tower.steam.active = tower.entity.is_crafting() and tower.entity.crafting_progress < 1 
    -- reset steam puff crafting progress so it never actually finishes
	tower.steam.crafting_progress = 0.1 
  end
end

end


-- HELPER FUNCTIONS
do

-- writes a message into the log file
function logging(message)
  	if settings.global["write-log"].value == true then
		game.write_file("RealisticReactors.log","\r\n[" .. game.tick .. "] " .. message,true)
	end
end

-- screen message to all players
function msg(s)
	for _, player in pairs(game.players) do
		if player.connected then
			player.print(s)
		end
	end
end

-- merges two tables
function union_tables(t1, t2)

   for i,v in ipairs(t2) do
      table.insert(t1, v)
   end 

   return t1
end

function table_length(tbl)
	if tbl == nil then
		return 0
	else
		local count = 0
		for _ in pairs(tbl) do
			count = count + 1
		end
		return count	
	end
end


end






